#include <vector>
#include <iostream>
#include <fstream>
#include <cmath>
#include <chrono>

#include "./revengmc.hpp"

#define MAX_48 0xFFFFFFFFFFFF
#define WORKER_COUNT 1024
#define MAGIC_NUM 25214903917

class noop_stream final : public std::ostream {
private:
	class buffer final : public std::streambuf {
	protected:
		[[nodiscard]] auto overflow(int_type ch) noexcept -> int_type override {
			return ch;
		}
	};

	buffer os_buffer_{};

public:
	noop_stream() noexcept : std::ostream(&os_buffer_) { }
};

condition_t condition_t::make(const feature_t &feat, int32_t x, int32_t y) {
	condition_t res;

	res.x = x >> 4;
	res.y = y >> 4;

	int32_t scaled_x = std::floor((double_t)res.x / feat.spacing);
	int32_t scaled_y = std::floor((double_t)res.y / feat.spacing);

	res.diff = feat.spacing - feat.separation;
	res.part_seed = (uint64_t)scaled_x * 341873128712L + (uint64_t)scaled_y * 132897987541L + (uint64_t)feat.salt;

	res.res_x = scaled_x * feat.spacing;
	res.res_y = scaled_y * feat.spacing;

	res.linear_sep = feat.linear_sep;

	return res;
}

const std::map<std::string, feature_t> FEATURES = {
	{ "MINESHAFT", { 1, 0, 0 } },
	{ "MINESHAFT_MESA", { 1, 0, 0 } },
	{ "MANSION", { 80, 20, 10387319, false } },
	{ "JUNGLE_PYRAMID", { 32, 8, 14357619 } },
	{ "DESERT_PYRAMID", { 32, 8, 14357617 } },
	{ "IGLOO", { 32, 8, 14357618 } },
	{ "SHIPWRECK", { 24, 4, 165745295 } },
	{ "SHIPWRECK_BEACHED", { 24, 4, 165745295 } },
	{ "SWAMP_HUT", { 32, 8, 14357620 } },
	{ "STRONGHOLD", { 1, 0, 0 } },
	{ "MONUMENT", { 32, 5, 10387313, false } },
	{ "OCEAN_RUIN_COLD", { 20, 8, 14357621 } },
	{ "OCEAN_RUIN_WARM", { 20, 8, 14357621 } },
	{ "FORTRESS", { 27, 4, 30084232 } },
	{ "NETHER_FOSSIL", { 2, 1, 14357921 } },
	{ "END_CITY", { 20, 11, 10387313, false } },
	{ "BURIED_TREASURE", { 1, 0, 0 } },
	{ "BASTION_REMNANT", { 27, 4, 30084232 } },
	{ "VILLAGE_PLAINS", { 32, 8, 10387312 } },
	{ "VILLAGE_DESERT", { 32, 8, 10387312 } },
	{ "VILLAGE_SAVANNA", { 32, 8, 10387312 } },
	{ "VILLAGE_SNOWY", { 32, 8, 10387312 } },
	{ "VILLAGE_TAIGA", { 32, 8, 10387312 } },
	{ "RUINED_PORTAL", { 40, 15, 34222645 } },
	{ "RUINED_PORTAL_DESERT", { 40, 15, 34222645 } },
	{ "RUINED_PORTAL_JUNGLE", { 40, 15, 34222645 } },
	{ "RUINED_PORTAL_SWAMP", { 40, 15, 34222645 } },
	{ "RUINED_PORTAL_MOUNTAIN", { 40, 15, 34222645 } },
	{ "RUINED_PORTAL_OCEAN", { 40, 15, 34222645 } },
	{ "RUINED_PORTAL_NETHER", { 40, 15, 34222645 } },
	{ "PILLAGER_OUTPOST", { 32, 8, 165745296 } },
};



size_t percent = 0;
bool running;

std::string read_file(const std::string &path) {
	std::ifstream f { path };
	if (f.fail()) return nullptr;

	f.seekg(0, std::ios_base::end);
	size_t n = f.tellg();
	f.seekg(0, std::ios_base::beg);

	auto buff = new char[n + 1];
	buff[n] = 0;
	f.read(buff, n);

	std::string res { buff };
	return res;
}

void read_input(std::istream &in, std::ostream &out, std::vector<condition_t> &tasks, size_t &start, size_t &end) {
	size_t task_n;

	std::string start_str, end_str;

	out << "Feature count: ";
	in >> task_n;

	out << "Seed range: ";
	in >> start_str >> end_str;

	if (start_str.find("0x") == 0) sscanf(start_str.c_str(), "%lx", &start);
	else sscanf(start_str.c_str(), "%lu", &start);

	if (end_str.find("0x") == 0) sscanf(end_str.c_str(), "%lx", &end);
	else sscanf(end_str.c_str(), "%lu", &end);

	for (size_t i = 0; i < task_n; i++) {
		std::string name;
		int32_t x, y;

		out << "Structure " << i + 1 << " name, X, Y: ";
		in >> name >> x >> y;

		auto it = FEATURES.find(name);
		if (it == FEATURES.end()) {
			std::cerr << "No structure named " << name << " found!\n";
			std::exit(1);
		}

		tasks.push_back(condition_t::make(it->second, x, y));
	}
}

std::string stringify_time(size_t time) {
	time /= 1000;
	int seconds = time % 60;
	int minutes = (time / 60) % 60;
	int hours = (time / 3600) % 24;
	int days = time / (3600 * 24);

	char res[256];
	snprintf(res, sizeof res, "%d:%.2d:%.2d:%.2d", days, hours, minutes, seconds);
	return std::string(res);
}

int main(int argc, char **argv) {
	std::vector<condition_t> tasks;
	size_t start, end;

	if (argc >= 2) {
		auto in = std::ifstream(argv[1]);
		auto out = noop_stream();
		if (in.fail()) {
			std::cerr << "Invalid input file provided!";
			return -1;
		}

		read_input(in, out, tasks, start, end);
	}
	else {
		read_input(std::cin, std::cerr, tasks, start, end);
	}

	engine::init(tasks);
	size_t batch_size = engine::batch_size();
	size_t elapsed = 0;

	for (size_t curr_start = start; curr_start <= end;) {
		auto time_start = std::chrono::high_resolution_clock::now();

		size_t curr_end = std::min(curr_start + batch_size, end + 1);

		size_t done = curr_start - start;
		size_t all = end - start + 1;
		float percent = (float)done / all * 100.f;

		size_t estimated = done == 0 ? 0 : ((size_t)(elapsed / (double)done * all));
		fprintf(
			stderr, "%.4f%%, remaining: %s, elapsed: %s, estimated: %s, %lX out of %lX done, current batch: %lX - %lX\n",
			percent,
			stringify_time(estimated - elapsed).c_str(), stringify_time(elapsed).c_str(), stringify_time(estimated).c_str(),
			done, all, curr_start, curr_end - 1
		);

		auto &seeds = engine::run(curr_start, curr_end);
		for (auto it = seeds.begin(); it != seeds.end(); it++) {
			fprintf(stderr, "SEED FOUND! %lu\n", *it);
		}

		curr_start = curr_end;

		auto time_end = std::chrono::high_resolution_clock::now();
		elapsed += std::chrono::duration_cast<std::chrono::milliseconds>(time_end - time_start).count();
	}

	engine::free();
}
